////////////////////////////////////////////////////////////////////////////////////////////
// CONSTANTS
////////////////////////////////////////////////////////////////////////////////////////////

const CALCULATOR_LOADED = true;
const CALCULATOR_NAME = "Calculator"
const CALCULATOR_INPUT = "calc_input"
const WINDOW = "window"
const WRONG = document.getElementById("calc-bundle").getString("calc.wrong");
const WRONG_PAR = document.getElementById("calc-bundle").getString("calc.wrongPar");
const WRONG_FUNC = document.getElementById("calc-bundle").getString("calc.wrongFunc");
const BASE = 10
const PRECISION = 12


////////////////////////////////////////////////////////////////////////////////////////////
// Class implementation
////////////////////////////////////////////////////////////////////////////////////////////

function Calculator() {
  this.init();
}

Calculator.prototype = {
  result : null,
  history : null,
  functions : null,
  mBranch : null,

  init : function() {
    try {
      this.history = new History(this);
      this.functions = new Functions(this);
      var input = document.getElementById(CALCULATOR_INPUT);
      input.focus();
      // Manage UI options
      var prefService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
      this.mBranch = prefService.getBranch("calculator.");
      var aButton = document.getElementById("show-hide-button");
      if (this.getPreference(PREF_NAME_CALC_PAD, "Int") == PAD_YES) {
	document.getElementById("calculator-buttons").hidden = false;
	aButton.image="chrome://calculator/skin/hide.png";
      }
      else {
	document.getElementById("calculator-buttons").hidden = true;
	aButton.image="chrome://calculator/skin/show.png";
      }
    } catch (e) {
      alert(e);
    }
  },

  close : function() {
  },

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Setters and getters
  ////////////////////////////////////////////////////////////////////////////////////////////

  setValue : function(id, value) {
    try {
      document.getElementById(id).value = value;
    } catch (e) {
    }
  },


  getValue : function(id) {
    try {
      return document.getElementById(id).value;
    } catch (e) {
    }
    return null;
  },
  getPreference : function(aPref, aType, aDefault) {
    try {
      var val = eval("this.mBranch.get" + aType + "Pref('" + aPref + "')");
      return val;
    } catch(e) {
      this.setPreference(aPref, aType, aDefault);
      return aDefault;
    }
    return null;
  },
  setPreference : function(aPref, aType, aValue) {
    try {
      eval("this.mBranch.set" + aType + "Pref('" + aPref + "'," + aValue + ")");
      return true;
    } catch(e) {
      alert(e);
    }
    return false;
  },


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Commands
  ////////////////////////////////////////////////////////////////////////////////////////////

  enter : function() {
    if (this.getValue(CALCULATOR_INPUT) != "") {
      this.history.addExpression(this.getValue(CALCULATOR_INPUT), this.evaluate(this.getValue(CALCULATOR_INPUT)));
    }
    this.reset();
//     var input = document.getElementById(CALCULATOR_INPUT);
//     input.focus();
  },

  reset : function() {
    this.setValue(CALCULATOR_INPUT, "");
    var input = document.getElementById(CALCULATOR_INPUT);
    input.focus();
  },

  exit : function() {
    window.close();
  },
  save : function() {
  },
  help : function() {
    window.open("help.xul", "Help", "chrome");
  },
  about : function() {
    window.openDialog("chrome://calculator/content/about.xul", "About", "chrome");
  },
  showKeys : function() {
    var element = document.getElementById("calculator-buttons");
    var aButton = document.getElementById("show-hide-button");
    this.history.scroll();
    element.hidden = ! element.hidden;
    if (element.hidden) {
      aButton.image="chrome://calculator/skin/show.png";
      this.setPreference(PREF_NAME_CALC_PAD, "Int", PAD_NO);
    } else {
      aButton.image="chrome://calculator/skin/hide.png";
      this.setPreference(PREF_NAME_CALC_PAD, "Int", PAD_YES);
    }
  },

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Interface related
  ////////////////////////////////////////////////////////////////////////////////////////////

  addToInput : function(func) {
    var input = document.getElementById(CALCULATOR_INPUT);
    var str = input.value;
    var curspos = input.selectionEnd;
    var res = "";
    var newPos = curspos + func.length;
    if (func.indexOf("(") != -1) {
      func += ")";
    }
    res = res.concat(str.slice(0, curspos), func, str.slice(curspos, str.length));
    input.value = res;
    input.setSelectionRange(newPos, newPos);
    input.focus();
  },

  checkInput : function() {
    var input = document.getElementById(CALCULATOR_INPUT);
    var str = input.value;
    if (str.length == 1) {
      var c = str[0];
      if ((c == '/' || c == '*' || c == '+' || c == '%' || c == '^') && this.history.index != 0) {
	this.history.current_pos = this.history.index - 1;
	this.history.setInputByKey();
	input.value = input.value.concat(c);
      }
    }
  },

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Core business
  ////////////////////////////////////////////////////////////////////////////////////////////

  basicEval : function(expr) {
    try {
      var i;
      for (i = 0; i < expr.length; i++) {
	var c = expr[i];
	if (c >= 'a' && c <= 'z')
	  return expr;
      }
      var num = new Number(eval(expr));
      num = num.toFixed(PRECISION);
      num = num.toString(BASE);
      if (num == "NaN") {
	num = 0;
      }
      return eval(num+0);
    }
    catch (e) {
      throw WRONG;
    }
  },

  findFunction : function(expr) {
    var i;
    var j;
    for (i = 0; i < expr.length; i++) {
      var c = expr[i];
      // Find functions
      if (!((c >= '0' && c <= '9') || c == '+' || c == '-' || c == '*' || c == '/' || c == '.'
	    || c == '(' || c == ')' || c == '%')) {

	// if it's a base > 10
	for (j = i; j < expr.length; j++) {
	  c = expr[j];
	  if (!((c >= 'a' && c <= 'z') || (c >= '0' && c <= '9'))){
	    break;
	  }
	}
	if (c != '(')
	  continue;

	// Extract the sub expression
	var subexpr = this.getSubExpr(expr, i);
	// Select the function and calculate
	var res = this.functions.selectFunction(subexpr);
	if (res == WRONG_FUNC) {
	  throw WRONG_FUNC;
	}
	// put the calculated value into expr
	expr = this.replaceSubstring(expr, subexpr, res);
      }
    }
    return this.basicEval(expr);
  },

  // This is kinda the main thing...
  evaluate : function (expr) {
    try {
      // Cleaning and replacing
      expr = this.formatStylePassIn(expr);
      expr = this.constantPass(expr);
      expr = expr.toLowerCase();
      expr = this.signPass(expr);
      // The real important line...
      return this.formatStylePassOut(this.findFunction(expr));
    }
    catch (e) {
      return e;
    }
    return WRONG;
  },

  formatStylePassIn : function(expr) {
    expr = expr.replace(new RegExp(' ', "gi"), '');
    if (this.getPreference(PREF_NAME_FORMAT, "Int", FORMAT_EN) == FORMAT_FR) {
      expr = expr.replace(new RegExp(',', "gi"), '.');
      expr = expr.replace(new RegExp(';', "gi"), ',');
    }
    return expr;
  },
  formatStylePassOut : function(expr) {
    //    expr = expr.toFixed(8);    // fix for a javascrit bug
    expr += "";
    if (this.getPreference(PREF_NAME_FORMAT, "Int", FORMAT_EN) == FORMAT_FR) {
      expr = expr.replace(new RegExp(',', "gi"), ';');
      expr = expr.replace(new RegExp('\\.', "gi"), ',');
    }
    return expr;
  },

  constantPass : function(expr) {
    expr = expr.replace(new RegExp('LOG10E', "gi"), Math.LOG10E.toString(BASE));
    expr = expr.replace(new RegExp('LOG2E', "gi"), Math.LOG2E.toString(BASE));
    expr = expr.replace(new RegExp('E', "g"), Math.E.toString(BASE));
    expr = expr.replace(new RegExp('LN10', "gi"), Math.LN10.toString(BASE));
    expr = expr.replace(new RegExp('LN2', "gi"), Math.LN2.toString(BASE));
    expr = expr.replace(new RegExp('PI', "gi"), Math.PI.toString(BASE));
    expr = expr.replace(new RegExp('SQRT1/2', "gi"), Math.SQRT1_2.toString(BASE));
    expr = expr.replace(new RegExp('SQRT2', "gi"), Math.SQRT2.toString(BASE));
    return expr;
  },

  signPass : function(expr) {
    var re = new RegExp('!([0-9.]+)', "gi");
    if (re.test(expr)) {
      expr = expr.replace(re, 'fact($1)');
    }
    re = RegExp("!", "gi");
    if (re.test(expr)) {
      expr = expr.replace(re, 'fact');
    }
//     re = RegExp("([0-9.]+)\\^([0-9.]+)", "gi");
//     if (re.test(expr)) {
//       expr = expr.replace(re, 'pow($1,$2)');
//     }
    return expr;
  },

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Utilities
  ////////////////////////////////////////////////////////////////////////////////////////////

  getSubExpr : function(expr, pos) {
    var par = -1;
    var i;
    for (i = pos; i < expr.length; i++) {
      if (expr[i] == '(') {
	if (par == -1) {
	  par = 0;
	}
	par++;
      }
      else if ( expr[i] == ')') {
	par--;
      }
      if (par == 0) {
	break;
      }
    }
    if (par != 0) {
      throw WRONG_PAR;
    }
    return expr.substring(pos, i + 1);
  },

  replaceSubstring : function(inputString, fromString, toString) {
    // Goes through the inputString and replaces every occurrence of fromString with toString
    var temp = inputString;
    if (fromString == "") {
      return inputString;
    }
    if (toString.indexOf(fromString) == -1) { // If the string being replaced is not a part of the replacement string (normal situation)
      while (temp.indexOf(fromString) != -1) {
	var toTheLeft = temp.substring(0, temp.indexOf(fromString));
	var toTheRight = temp.substring(temp.indexOf(fromString)+fromString.length, temp.length);
	temp = toTheLeft + toString + toTheRight;
      }
    } else { // String being replaced is part of replacement string (like "+" being replaced with "++") - prevent an infinite loop
      var midStrings = new Array("~", "`", "_", "^", "#");
      var midStringLen = 1;
      var midString = "";
      // Find a string that doesn't exist in the inputString to be used
      // as an "inbetween" string
      while (midString == "") {
	for (var i=0; i < midStrings.length; i++) {
	  var tempMidString = "";
	  for (var j=0; j < midStringLen; j++) { tempMidString += midStrings[i]; }
	  if (fromString.indexOf(tempMidString) == -1) {
	    midString = tempMidString;
	    i = midStrings.length + 1;
	  }
	}
      } // Keep on going until we build an "inbetween" string that doesn't exist
      // Now go through and do two replaces - first, replace the "fromString" with the "inbetween" string
      while (temp.indexOf(fromString) != -1) {
	toTheLeft = temp.substring(0, temp.indexOf(fromString));
	toTheRight = temp.substring(temp.indexOf(fromString)+fromString.length, temp.length);
	temp = toTheLeft + midString + toTheRight;
      }
      // Next, replace the "inbetween" string with the "toString"
      while (temp.indexOf(midString) != -1) {
	toTheLeft = temp.substring(0, temp.indexOf(midString));
	toTheRight = temp.substring(temp.indexOf(midString)+midString.length, temp.length);
	temp = toTheLeft + toString + toTheRight;
      }
    } // Ends the check to see if the string being replaced is part of the replacement string or not
    return temp; // Send the updated string back to the user
  } // Ends the "replaceSubstring" function
}
////////////////////////////////////////////////////////////////////////////////////////////
// The end
////////////////////////////////////////////////////////////////////////////////////////////
